"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.ExpressionHelper = void 0;
const generated_1 = require("../generated");
const ExtendedCircomVisitor_1 = require("../ExtendedCircomVisitor");
class ExpressionHelper {
    templateIdentifier;
    _expressionContext = null;
    _variableContext = {};
    constructor(templateIdentifier) {
        this.templateIdentifier = templateIdentifier;
    }
    setExpressionContext(expressionContext) {
        this._expressionContext = expressionContext;
        return this;
    }
    setVariableContext(variableContext) {
        this._variableContext = variableContext;
        return this;
    }
    addVariablesToTheContext(variables) {
        for (const key in variables) {
            this._variableContext[key] = variables[key];
        }
        return this;
    }
    /**
     * This function can throw an error if the expression context is not set.
     * All other errors related to the parsing of the expression can be retrieved by the function `getErrors()`
     * inside the `ExtendedCircomVisitor` class.
     */
    parseExpression() {
        if (!this._expressionContext) {
            throw new Error("Expression context is not set");
        }
        const visitor = new ExpressionVisitor(this.templateIdentifier, this._variableContext);
        const result = visitor.visitExpression(this._expressionContext);
        if (result === null) {
            return [null, visitor.getErrors()];
        }
        return [result, visitor.getErrors()];
    }
}
exports.ExpressionHelper = ExpressionHelper;
class ExpressionVisitor extends ExtendedCircomVisitor_1.ExtendedCircomVisitor {
    variableContext;
    constructor(templateIdentifier, variableContext) {
        super(templateIdentifier);
        this.variableContext = variableContext;
    }
    visitExpression = (ctx) => {
        if (ctx.PARALLEL()) {
            return this.visit(ctx.expression(0));
        }
        if (ctx.primaryExpression()) {
            return this.visit(ctx.primaryExpression());
        }
        if (ctx.TERNARY_CONDITION() && ctx.TERNARY_ALTERNATIVE()) {
            const conditionResult = this.visit(ctx._cond);
            if (conditionResult === null) {
                this.addError("Failed to resolve the condition of a ternary expression", ctx._cond);
                return null;
            }
            if (Array.isArray(conditionResult)) {
                this.addError("Ternary expression condition cannot be an array", ctx._cond);
                return null;
            }
            if (conditionResult === 1n) {
                return this.visit(ctx._ifTrue);
            }
            else {
                return this.visit(ctx._ifFalse);
            }
        }
        return this._parseOperation(ctx);
    };
    visitPIdentifierStatement = (ctx) => {
        if (ctx.identifierStatement().idetifierAccess_list().length == 0) {
            const variableName = this.variableContext[ctx.identifierStatement().ID().getText()];
            if (variableName === undefined) {
                this.addError(`Variable ${ctx.identifierStatement().ID().getText()} is not defined`, ctx.identifierStatement());
                return null;
            }
            return variableName;
        }
        this.addError("IdentifierStatement is not supported with access references", ctx);
        return null;
    };
    visitPUnderscore = (_ctx) => {
        return null;
    };
    visitPNumber = (ctx) => {
        return BigInt(ctx.NUMBER().getText());
    };
    visitPParentheses = (ctx) => {
        this.addError("Parentheses are not supported", ctx);
        return null;
    };
    visitPArray = (ctx) => {
        let arrayItems = [];
        for (let i = 0; i < ctx.expressionList().expression_list().length; i++) {
            const resolvedItem = this.visit(ctx.expressionList().expression(i));
            if (!resolvedItem) {
                this.addError(`Failed to resolve the ${i} element of an array.`, ctx.expressionList().expression(i));
                arrayItems = null;
                continue;
            }
            if (arrayItems) {
                arrayItems.push(resolvedItem);
            }
        }
        return arrayItems;
    };
    visitPCall = (ctx) => {
        this.addError("Calls are not supported", ctx);
        return null;
    };
    visitPAnonymousCall = (ctx) => {
        this.addError("Anonymous calls are not supported", ctx);
        return null;
    };
    _parseOperation(ctx) {
        const operationType = ctx._op.type;
        const firstExpression = this.visit(ctx.expression(0));
        if (firstExpression === null) {
            this.addError("Failed to resolve the first expression of an operation", ctx.expression(0));
            return null;
        }
        if (Array.isArray(firstExpression)) {
            this.addError("Performing operations on arrays is not allowed", ctx.expression(0));
            return null;
        }
        switch (operationType) {
            case generated_1.CircomParser.NOT:
                if (firstExpression !== 0n && firstExpression !== 1n) {
                    this.addError("NOT operation can be performed only on boolean values", ctx.expression(0));
                    return null;
                }
                return firstExpression ? 0n : 1n;
            case generated_1.CircomParser.BNOT:
                return ~firstExpression;
            case generated_1.CircomParser.SUB:
                return -firstExpression;
        }
        const secondExpression = this.visit(ctx.expression(1));
        if (secondExpression === null) {
            this.addError("Failed to resolve the second expression of an operation", ctx.expression(1));
            return null;
        }
        if (Array.isArray(secondExpression)) {
            this.addError("Performing operations on arrays is not allowed", ctx.expression(1));
            return null;
        }
        switch (operationType) {
            case generated_1.CircomParser.POW:
                return firstExpression ** secondExpression;
            case generated_1.CircomParser.MUL:
                return firstExpression * secondExpression;
            case generated_1.CircomParser.DIV:
                return firstExpression / secondExpression;
            case generated_1.CircomParser.QUO:
                this.addError("QUO operation is not supported", ctx);
                return null;
            case generated_1.CircomParser.MOD:
                return firstExpression % secondExpression;
            case generated_1.CircomParser.ADD:
                return firstExpression + secondExpression;
            case generated_1.CircomParser.SUB:
                return firstExpression - secondExpression;
            case generated_1.CircomParser.SHL:
                return firstExpression << secondExpression;
            case generated_1.CircomParser.SHR:
                return firstExpression >> secondExpression;
            case generated_1.CircomParser.BAND:
                return firstExpression & secondExpression;
            case generated_1.CircomParser.BXOR:
                return firstExpression ^ secondExpression;
            case generated_1.CircomParser.BOR:
                return firstExpression | secondExpression;
            case generated_1.CircomParser.EQ:
                return firstExpression === secondExpression ? 1n : 0n;
            case generated_1.CircomParser.NEQ:
                return firstExpression !== secondExpression ? 1n : 0n;
            case generated_1.CircomParser.LT:
                return firstExpression < secondExpression ? 1n : 0n;
            case generated_1.CircomParser.GT:
                return firstExpression > secondExpression ? 1n : 0n;
            case generated_1.CircomParser.LE:
                return firstExpression <= secondExpression ? 1n : 0n;
            case generated_1.CircomParser.GE:
                return firstExpression >= secondExpression ? 1n : 0n;
            case generated_1.CircomParser.AND:
                return firstExpression && secondExpression ? 1n : 0n;
            case generated_1.CircomParser.OR:
                return firstExpression || secondExpression ? 1n : 0n;
        }
        this.addError("Reached unknown operation", ctx);
        return null;
    }
}
